from typing import Union
from mpi4py.MPI import Intracomm

__version__: str = ...
def init(args: Union[str, list[str]] | None = ..., arch: str | None = ..., comm: Intracomm | None = ...) -> None: ...
def get_include() -> str: ...
def get_config() -> dict[str, str]: ...
